@@ -85,8 +85,8 @@ static void setup_access_to_local_objects(struct request *req)
 	copies = hdr->copies;
 	if (!copies)
 		copies = sys->nr_sobjs;
-	if (copies > req->nr_nodes)
-		copies = req->nr_nodes;
+	if (copies > req->nr_zones)
+		copies = req->nr_zones;
 
 	if (is_access_local(req->entry, req->nr_vnodes, hdr->oid, copies))
 		req->local_oid = hdr->oid;
@@ -99,8 +99,8 @@ static void __done(struct work *work, int idx)
 	int again = 0;
 	int copies = sys->nr_sobjs;
 
-	if (copies > req->nr_nodes)
-		copies = req->nr_nodes;
+	if (copies > req->nr_zones)
+		copies = req->nr_zones;
 
 	switch (hdr->opcode) {
 	case SD_OP_NEW_VDI:
@@ -143,11 +143,18 @@ static void __done(struct work *work, int idx)
 		} else if (req->rp.result == SD_RES_SUCCESS && req->check_consistency) {
 			struct sd_obj_req *obj_hdr = (struct sd_obj_req *)&req->rq;
 			uint32_t vdi_id = oid_to_vid(obj_hdr->oid);
-			struct data_object_bmap *bmap;
+			struct data_object_bmap *bmap, *n;
+			int nr_bmaps = 0;
 
-			list_for_each_entry(bmap, &sys->consistent_obj_list, list) {
+			if (!is_data_obj(obj_hdr->oid))
+				goto done;
+
+			list_for_each_entry_safe(bmap, n, &sys->consistent_obj_list, list) {
+				nr_bmaps++;
 				if (bmap->vdi_id == vdi_id) {
 					set_bit(data_oid_to_idx(obj_hdr->oid), bmap->dobjs);
+					list_del(&bmap->list);
+					list_add_tail(&bmap->list, &sys->consistent_obj_list);
 					goto done;
 				}
 			}
@@ -158,8 +165,15 @@ static void __done(struct work *work, int idx)
 			}
 			dprintf("allocate a new object map\n");
 			bmap->vdi_id = vdi_id;
-			list_add(&bmap->list, &sys->consistent_obj_list);
+			list_add_tail(&bmap->list, &sys->consistent_obj_list);
 			set_bit(data_oid_to_idx(obj_hdr->oid), bmap->dobjs);
+			if (nr_bmaps >= MAX_DATA_OBJECT_BMAPS) {
+				/* the first entry is the least recently used one */
+				bmap = list_first_entry(&sys->consistent_obj_list,
+							struct data_object_bmap, list);
+				list_del(&bmap->list);
+				free(bmap);
+			}
 		} else if (is_access_local(req->entry, req->nr_vnodes,
 					   ((struct sd_obj_req *)&req->rq)->oid, copies) &&
 			   req->rp.result == SD_RES_EIO) {
@@ -194,6 +208,8 @@ static void queue_request(struct request *req)
 	struct sd_req *hdr = (struct sd_req *)&req->rq;
 	struct sd_rsp *rsp = (struct sd_rsp *)&req->rp;
 
+	dprintf("%x\n", hdr->opcode);
+
 	if (hdr->opcode == SD_OP_KILL_NODE) {
 		log_close();
 		exit(1);
@@ -255,6 +271,7 @@ static void queue_request(struct request *req)
 	case SD_OP_SHUTDOWN:
 	case SD_OP_STAT_CLUSTER:
 	case SD_OP_GET_VDI_ATTR:
+	case SD_OP_GET_EPOCH:
 		req->work.fn = cluster_queue_request;
 		break;
 	case SD_OP_READ_VDIS:
@@ -331,16 +348,18 @@ static void free_request(struct request *req)
 static void req_done(struct request *req)
 {
 	int dead = 0;
+	struct client_info *ci = req->ci;
 
-	if (conn_tx_on(&req->ci->conn)) {
+	if (conn_tx_on(&ci->conn)) {
 		dprintf("connection seems to be dead\n");
 		dead = 1;
 	} else
-		list_add(&req->r_wlist, &req->ci->done_reqs);
-	client_decref(req->ci);
+		list_add(&req->r_wlist, &ci->done_reqs);
 
 	if (dead)
 		free_request(req);
+
+	client_decref(ci);
 }
 
 static void init_rx_hdr(struct client_info *ci)
@@ -538,7 +557,7 @@ static void client_handler(int fd, int events, void *data)
 	if (events & EPOLLIN)
 		client_rx_handler(ci);
 
-	if (!is_conn_dead(&ci->conn) && events & EPOLLOUT)
+	if (events & EPOLLOUT)
 		client_tx_handler(ci);
 
 	if (is_conn_dead(&ci->conn)) {
@@ -606,7 +625,7 @@ int create_listen_port(int port, void *data)
 }
 
 int write_object(struct sheepdog_vnode_list_entry *e,
-		 int vnodes, int nodes, uint32_t node_version,
+		 int vnodes, int zones, uint32_t node_version,
 		 uint64_t oid, char *data, unsigned int datalen,
 		 uint64_t offset, int nr, int create)
 {
@@ -614,16 +633,15 @@ int write_object(struct sheepdog_vnode_list_entry *e,
 	int i, n, fd, ret, success = 0;
 	char name[128];
 
-	if (nr > nodes)
-		nr = nodes;
+	if (nr > zones)
+		nr = zones;
 
 	for (i = 0; i < nr; i++) {
 		unsigned rlen = 0, wlen = datalen;
 
 		n = obj_to_sheep(e, vnodes, oid, i);
 
-		if (memcmp(e[n].addr, sys->this_node.addr, sizeof(e[n].addr)) == 0 &&
-		    e[n].port == sys->this_node.port) {
+		if (is_myself(e[n].addr, e[n].port)) {
 			ret = write_object_local(oid, data, datalen, offset, nr,
 						 node_version, create);
 
@@ -669,7 +687,7 @@ int write_object(struct sheepdog_vnode_list_entry *e,
 }
 
 int read_object(struct sheepdog_vnode_list_entry *e,
-		int vnodes, int nodes, uint32_t node_version,
+		int vnodes, int zones, uint32_t node_version,
 		uint64_t oid, char *data, unsigned int datalen,
 		uint64_t offset, int nr)
 {
@@ -678,15 +696,14 @@ int read_object(struct sheepdog_vnode_list_entry *e,
 	char name[128];
 	int i = 0, n, fd, ret, last_error = SD_RES_SUCCESS;
 
-	if (nr > nodes)
-		nr = nodes;
+	if (nr > zones)
+		nr = zones;
 
 	/* search a local object first */
 	for (i = 0; i < nr; i++) {
 		n = obj_to_sheep(e, vnodes, oid, i);
 
-		if (memcmp(e[n].addr, sys->this_node.addr, sizeof(e[n].addr)) == 0 &&
-		    e[n].port == sys->this_node.port) {
+		if (is_myself(e[n].addr, e[n].port)) {
 			ret = read_object_local(oid, data, datalen, offset, nr,
 						node_version);
 
@@ -741,7 +758,7 @@ int read_object(struct sheepdog_vnode_list_entry *e,
 }
 
 int remove_object(struct sheepdog_vnode_list_entry *e,
-		  int vnodes, int nodes, uint32_t node_version,
+		  int vnodes, int zones, uint32_t node_version,
 		  uint64_t oid, int nr)
 {
 	char name[128];
@@ -749,8 +766,8 @@ int remove_object(struct sheepdog_vnode_list_entry *e,
 	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&hdr;
 	int i = 0, n, fd, ret;
 
-	if (nr > nodes)
-		nr = nodes;
+	if (nr > zones)
+		nr = zones;
 
 	for (i = 0; i < nr; i++) {
 		unsigned wlen = 0, rlen = 0;
@@ -789,14 +806,14 @@ int remove_object(struct sheepdog_vnode_list_entry *e,
 int get_sheep_fd(uint8_t *addr, uint16_t port, int node_idx,
 		 uint32_t epoch, int worker_idx)
 {
-	static int cached_fds[NR_WORKER_THREAD][SD_MAX_NODES];
+	static int cached_fds[NR_GW_WORKER_THREAD][SD_MAX_NODES];
 	static uint32_t cached_epoch = 0;
 	int i, j, fd, ret;
 	char name[INET6_ADDRSTRLEN];
 
 	if (cached_epoch == 0) {
 		/* initialize */
-		for (i = 0; i < NR_WORKER_THREAD; i++) {
+		for (i = 0; i < NR_GW_WORKER_THREAD; i++) {
 			for (j = 0; j < SD_MAX_NODES; j++)
 				cached_fds[i][j] = -1;
 		}
@@ -809,7 +826,7 @@ int get_sheep_fd(uint8_t *addr, uint16_t port, int node_idx,
 		return -1;
 	}
 	if (after(epoch, cached_epoch)) {
-		for (i = 0; i < NR_WORKER_THREAD; i++) {
+		for (i = 0; i < NR_GW_WORKER_THREAD; i++) {
 			for (j = 0; j < SD_MAX_NODES; j++) {
 				if (cached_fds[i][j] >= 0)
 					close(cached_fds[i][j]);
