@@ -24,11 +24,13 @@
 #define SD_OP_REMOVE_OBJ     0x91
 
 #define SD_OP_GET_OBJ_LIST   0xA1
+#define SD_OP_GET_EPOCH      0XA2
 
 #define SD_MSG_JOIN             0x01
 #define SD_MSG_VDI_OP           0x02
 #define SD_MSG_MASTER_CHANGED   0x03
 #define SD_MSG_LEAVE            0x04
+#define SD_MSG_MASTER_TRANSFER  0x05
 
 #define SD_STATUS_OK                0x00
 #define SD_STATUS_WAIT_FOR_FORMAT   0x01
@@ -83,13 +85,15 @@ struct request {
 
 	struct sheepdog_vnode_list_entry entry[SD_MAX_VNODES];
 	int nr_vnodes;
-	int nr_nodes;
+	int nr_zones;
 	int check_consistency;
 
 	req_end_t done;
 	struct work work;
 };
 
+#define MAX_DATA_OBJECT_BMAPS 64
+
 struct data_object_bmap {
 	uint32_t vdi_id;
 	DECLARE_BITMAP(dobjs, MAX_DATA_OBJS);
@@ -115,6 +119,11 @@ struct cluster_info {
 	struct list_head cpg_node_list;
 	struct list_head sd_node_list;
 
+	/* leave list is only used to account for bad nodes when we start
+	 * up the cluster nodes after we shutdown the cluster through collie.
+	 */
+	struct list_head leave_list;
+
 	/* this array contains a list of ordered virtual nodes */
 	struct sheepdog_vnode_list_entry vnodes[SD_MAX_VNODES];
 	int nr_vnodes;
@@ -138,6 +147,12 @@ struct cluster_info {
 	uint32_t recovered_epoch;
 
 	int use_directio;
+
+	struct work_queue *cpg_wqueue;
+	struct work_queue *gateway_wqueue;
+	struct work_queue *io_wqueue;
+	struct work_queue *deletion_wqueue;
+	struct work_queue *recovery_wqueue;
 };
 
 extern struct cluster_info *sys;
@@ -146,6 +161,7 @@ int create_listen_port(int port, void *data);
 
 int is_io_request(unsigned op);
 int init_store(const char *dir);
+int init_base_path(const char *dir);
 
 int add_vdi(uint32_t epoch, char *data, int data_len, uint64_t size,
 	    uint32_t *new_vid, uint32_t base_vid, uint32_t copies,
@@ -160,19 +176,19 @@ int lookup_vdi(uint32_t epoch, char *data, int data_len, uint32_t *vid,
 int read_vdis(char *data, int len, unsigned int *rsp_len);
 
 int get_vdi_attr(uint32_t epoch, char *data, int data_len, uint32_t vid,
-		 uint32_t *attrid, int creat, int excl);
+		 uint32_t *attrid, int copies, int creat, int excl);
 
 int get_ordered_sd_node_list(struct sheepdog_node_list_entry *entries);
 void setup_ordered_sd_vnode_list(struct request *req);
 void get_ordered_sd_vnode_list(struct sheepdog_vnode_list_entry *entries,
-			       int *nr_vnodes, int *nr_nodes);
+			       int *nr_vnodes, int *nr_zones);
 int is_access_to_busy_objects(uint64_t oid);
 int is_access_local(struct sheepdog_vnode_list_entry *e, int nr_nodes,
 		    uint64_t oid, int copies);
 
 void resume_pending_requests(void);
 
-int create_cluster(int port);
+int create_cluster(int port, int64_t zone);
 int leave_cluster(void);
 
 void start_cpg_event_work(void);
@@ -191,10 +207,12 @@ int update_epoch_store(uint32_t epoch);
 int set_global_nr_copies(uint32_t copies);
 int get_global_nr_copies(uint32_t *copies);
 
-#define NR_WORKER_THREAD 64
+#define NR_GW_WORKER_THREAD 4
+#define NR_IO_WORKER_THREAD 4
 
 int epoch_log_write(uint32_t epoch, char *buf, int len);
 int epoch_log_read(uint32_t epoch, char *buf, int len);
+int epoch_log_read_remote(uint32_t epoch, char *buf, int len);
 int get_latest_epoch(void);
 int remove_epoch(int epoch);
 int set_cluster_ctime(uint64_t ctime);
@@ -205,15 +223,15 @@ void resume_recovery_work(void);
 int is_recoverying_oid(uint64_t oid);
 
 int write_object(struct sheepdog_vnode_list_entry *e,
-		 int vnodes, int nodes, uint32_t node_version,
+		 int vnodes, int zones, uint32_t node_version,
 		 uint64_t oid, char *data, unsigned int datalen,
 		 uint64_t offset, int nr, int create);
 int read_object(struct sheepdog_vnode_list_entry *e,
-		int vnodes, int nodes, uint32_t node_version,
+		int vnodes, int zones, uint32_t node_version,
 		uint64_t oid, char *data, unsigned int datalen,
 		uint64_t offset, int nr);
 int remove_object(struct sheepdog_vnode_list_entry *e,
-		  int vnodes, int nodes, uint32_t node_version,
+		  int vnodes, int zones, uint32_t node_version,
 		  uint64_t oid, int nr);
 
 int get_sheep_fd(uint8_t *addr, uint16_t port, int node_idx,
@@ -230,6 +248,7 @@ int get_sheep_fd(uint8_t *addr, uint16_t port, int node_idx,
 /* Journal header for data object */
 struct jrnl_vdi_head {
 	uint32_t jh_type;
+	uint32_t pad;
 	uint64_t jh_offset;
 	uint64_t jh_size;
 };
@@ -286,10 +305,4 @@ static inline int is_myself(uint8_t *addr, uint16_t port)
 		port == sys->this_node.port;
 }
 
-#define panic(fmt, args...)			\
-({						\
-	vprintf(SDOG_EMERG fmt, ##args);	\
-	exit(1);				\
-})
-
 #endif
