@@ -107,6 +107,10 @@ static int stat_sheep(uint64_t *store_size, uint64_t *store_free, uint32_t epoch
 	return SD_RES_SUCCESS;
 }
 
+static int merge_objlist(struct sheepdog_vnode_list_entry *entries, int nr_entries,
+			 uint64_t *list1, int nr_list1,
+			 uint64_t *list2, int nr_list2, int nr_objs);
+
 static int get_obj_list(struct request *req)
 {
 	DIR *dir;
@@ -114,12 +118,12 @@ static int get_obj_list(struct request *req)
 	struct sd_list_req *hdr = (struct sd_list_req *)&req->rq;
 	struct sd_list_rsp *rsp = (struct sd_list_rsp *)&req->rp;
 	uint64_t oid;
-	uint32_t epoch = hdr->tgt_epoch;
+	uint32_t epoch;
 	char path[1024];
 	uint64_t *p = (uint64_t *)req->data;
 	int nr = 0;
 	uint64_t *objlist = NULL;
-	int obj_nr = 0, fd, i;
+	int obj_nr, i;
 	int res = SD_RES_SUCCESS;
 	int buf_len;
 	char *buf;
@@ -133,65 +137,34 @@ static int get_obj_list(struct request *req)
 		goto out;
 	}
 
-	if (epoch == 1)
-		goto local;
-
-	snprintf(path, sizeof(path), "%s%08u/list", obj_path, epoch);
-
-	fd = open(path, O_RDONLY);
-	if (fd < 0) {
-		eprintf("failed to open %s, %s\n", path, strerror(errno));
-		res = SD_RES_EIO;
-		goto out;
-	}
-	obj_nr = read(fd, buf, buf_len);
-	dprintf("read objct list from %s, %"PRIu32"\n", path, obj_nr);
-
-	obj_nr /= sizeof(uint64_t);
 	objlist = (uint64_t *)buf;
-	for (i = 0; i < obj_nr; i++) {
-		p[nr++] = objlist[i];
+	for (epoch = 1; epoch <= hdr->tgt_epoch; epoch++) {
+		snprintf(path, sizeof(path), "%s%08u/", obj_path, epoch);
 
-		if (nr * sizeof(uint64_t) >= hdr->data_length)
-			break;
-	}
-	close(fd);
+		dprintf("%"PRIu32", %s\n", sys->this_node.port, path);
 
-local:
-	snprintf(path, sizeof(path), "%s%08u/", obj_path, hdr->tgt_epoch);
-
-	dprintf("%"PRIu32", %s\n", sys->this_node.port, path);
-
-	dir = opendir(path);
-	if (!dir) {
-		eprintf("%s\n", path);
-		res = SD_RES_EIO;
-		goto out;
-	}
-
-	while ((d = readdir(dir))) {
-		if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, ".."))
+		dir = opendir(path);
+		if (!dir) {
+			eprintf("%s\n", path);
 			continue;
+		}
 
-		oid = strtoull(d->d_name, NULL, 16);
-		if (oid == 0)
-			continue;
+		obj_nr = 0;
+		while ((d = readdir(dir))) {
+			if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, ".."))
+				continue;
 
-		for (i = 0; i < obj_nr; i++)
-			if (objlist[i] == oid)
-				break;
-		if (i < obj_nr)
-			continue;
+			oid = strtoull(d->d_name, NULL, 16);
+			if (oid == 0)
+				continue;
 
-		p[nr++] = oid;
+			objlist[obj_nr++] = oid;
+		}
 
-		if (nr * sizeof(uint64_t) >= hdr->data_length)
-			break;
+		closedir(dir);
+
+		nr = merge_objlist(NULL, 0, p, nr, objlist, obj_nr, 0);
 	}
-
-	eprintf("nr = %"PRIu32"\n", nr);
-
-	closedir(dir);
 out:
 	free(buf);
 	rsp->data_length = nr * sizeof(uint64_t);
@@ -308,8 +281,8 @@ static int forward_read_obj_req(struct request *req, int idx)
 	/* temporary hack */
 	if (!copies)
 		copies = sys->nr_sobjs;
-	if (copies > req->nr_nodes)
-		copies = req->nr_nodes;
+	if (copies > req->nr_zones)
+		copies = req->nr_zones;
 
 	hdr.flags |= SD_FLAG_CMD_DIRECT;
 
@@ -368,8 +341,8 @@ static int forward_write_obj_req(struct request *req, int idx)
 	/* temporary hack */
 	if (!copies)
 		copies = sys->nr_sobjs;
-	if (copies > req->nr_nodes)
-		copies = req->nr_nodes;
+	if (copies > req->nr_zones)
+		copies = req->nr_zones;
 
 	nr_fds = 0;
 	memset(pfds, 0, sizeof(pfds));
@@ -519,10 +492,14 @@ int update_epoch_store(uint32_t epoch)
 int write_object_local(uint64_t oid, char *data, unsigned int datalen,
 		       uint64_t offset, int copies, uint32_t epoch, int create)
 {
-	struct request req;
-	struct sd_obj_req *hdr = (struct sd_obj_req *)&req.rq;
+	int ret;
+	struct request *req;
+	struct sd_obj_req *hdr;
 
-	memset(&req, 0, sizeof(req));
+	req = zalloc(sizeof(*req));
+	if (!req)
+		return SD_RES_NO_MEM;
+	hdr = (struct sd_obj_req *)&req->rq;
 
 	hdr->oid = oid;
 	if (create)
@@ -533,20 +510,29 @@ int write_object_local(uint64_t oid, char *data, unsigned int datalen,
 	hdr->flags = SD_FLAG_CMD_WRITE;
 	hdr->offset = offset;
 	hdr->data_length = datalen;
-	req.data = data;
+	req->data = data;
 
-	return store_queue_request_local(&req, epoch);
+	ret = store_queue_request_local(req, epoch);
+
+	free(req);
+
+	return ret;
 }
 
 int read_object_local(uint64_t oid, char *data, unsigned int datalen,
 		      uint64_t offset, int copies, uint32_t epoch)
 {
 	int ret;
-	struct request req;
-	struct sd_obj_req *hdr = (struct sd_obj_req *)&req.rq;
-	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&req.rp;
+	struct request *req;
+	struct sd_obj_req *hdr;
+	struct sd_obj_rsp *rsp;
+	unsigned int rsp_data_length;
 
-	memset(&req, 0, sizeof(req));
+	req = zalloc(sizeof(*req));
+	if (!req)
+		return -SD_RES_NO_MEM;
+	hdr = (struct sd_obj_req *)&req->rq;
+	rsp = (struct sd_obj_rsp *)&req->rp;
 
 	hdr->oid = oid;
 	hdr->opcode = SD_OP_READ_OBJ;
@@ -554,17 +540,20 @@ int read_object_local(uint64_t oid, char *data, unsigned int datalen,
 	hdr->flags = 0;
 	hdr->offset = offset;
 	hdr->data_length = datalen;
-	req.data = data;
+	req->data = data;
 
-	ret = store_queue_request_local(&req, epoch);
+	ret = store_queue_request_local(req, epoch);
+
+	rsp_data_length = rsp->data_length;
+	free(req);
 
 	if (ret != 0)
 		return -ret;
 
-	if (rsp->data_length != datalen)
+	if (rsp_data_length != datalen)
 		return -SD_RES_EIO;
 
-	return rsp->data_length;
+	return rsp_data_length;
 }
 
 static int store_queue_request_local(struct request *req, uint32_t epoch)
@@ -579,6 +568,8 @@ static int store_queue_request_local(struct request *req, uint32_t epoch)
 	struct jrnl_descriptor jd;
 	struct jrnl_vdi_head jh;
 
+	dprintf("%x, %" PRIx64" , %u\n", opcode, oid, epoch);
+
 	switch (opcode) {
 	case SD_OP_CREATE_AND_WRITE_OBJ:
 	case SD_OP_WRITE_OBJ:
@@ -703,14 +694,20 @@ static int store_queue_request_local(struct request *req, uint32_t epoch)
 		break;
 	case SD_OP_WRITE_OBJ:
 	case SD_OP_CREATE_AND_WRITE_OBJ:
-		if (hdr->flags & SD_FLAG_CMD_TRUNCATE)
-			ftruncate(fd, hdr->offset + hdr->data_length);
+		if (hdr->flags & SD_FLAG_CMD_TRUNCATE) {
+			ret = ftruncate(fd, hdr->offset + hdr->data_length);
+			if (ret) {
+				ret = SD_RES_EIO;
+				goto out;
+			}
+		}
 
 		if (is_vdi_obj(oid)) {
 			jd.jdf_epoch = epoch;
 			jd.jdf_oid = oid;
 			jd.jdf_target_fd = fd;
 
+			memset(&jh, 0, sizeof(jh));
 			jh.jh_type = JRNL_TYPE_VDI;
 			jh.jh_offset = hdr->offset;
 			jh.jh_size = hdr->data_length;
@@ -756,7 +753,7 @@ out:
 
 static int fix_object_consistency(struct request *req, int idx)
 {
-	int ret;
+	int ret = SD_RES_NO_MEM;
 	unsigned int data_length;
 	struct sd_obj_req *hdr = (struct sd_obj_req *)&req->rq;
 	struct sd_obj_req req_bak = *((struct sd_obj_req *)&req->rq);
@@ -878,6 +875,50 @@ int epoch_log_write(uint32_t epoch, char *buf, int len)
 	return 0;
 }
 
+int epoch_log_read_remote(uint32_t epoch, char *buf, int len)
+{
+	struct sd_obj_req hdr;
+	struct sd_obj_rsp *rsp = (struct sd_obj_rsp *)&hdr;
+	int fd, i, ret;
+	unsigned int rlen, wlen, nr, le = get_latest_epoch();
+	char host[128];
+	struct sheepdog_node_list_entry nodes[SD_MAX_NODES];
+
+	nr = epoch_log_read(le, (char *)nodes, ARRAY_SIZE(nodes));
+	nr /= sizeof(nodes[0]);
+	for (i = 0; i < nr; i++) {
+		if (is_myself(nodes[i].addr, nodes[i].port))
+			continue;
+
+		addr_to_str(host, sizeof(host), nodes[i].addr, 0);
+		fd = connect_to(host, nodes[i].port);
+		if (fd < 0) {
+			vprintf(SDOG_ERR "can't connect to %s, %m\n", host);
+			continue;
+		}
+
+		memset(&hdr, 0, sizeof(hdr));
+		hdr.opcode = SD_OP_GET_EPOCH;
+		hdr.tgt_epoch = epoch;
+		hdr.data_length = len;
+		rlen = hdr.data_length;
+		wlen = 0;
+
+		ret = exec_req(fd, (struct sd_req *)&hdr, buf, &wlen, &rlen);
+		close(fd);
+
+		if (ret)
+			continue;
+		if (rsp->result == SD_RES_SUCCESS) {
+			ret = rsp->data_length;
+			goto out;
+		}
+	}
+	ret = 0; /* If no one has targeted epoch file, we can safely return 0 */
+out:
+	return ret;
+}
+
 int epoch_log_read(uint32_t epoch, char *buf, int len)
 {
 	int fd;
@@ -902,8 +943,10 @@ int get_latest_epoch(void)
 	uint32_t e, epoch = 0;
 
 	dir = opendir(epoch_path);
-	if (!dir)
-		return -1;
+	if (!dir) {
+		vprintf(SDOG_EMERG "failed to get the latest epoch, %m\n");
+		abort();
+	}
 
 	while ((d = readdir(dir))) {
 		e = atoi(d->d_name);
@@ -1020,6 +1063,27 @@ uint64_t get_cluster_ctime(void)
 	return ctime;
 }
 
+static int get_max_copies(struct sheepdog_node_list_entry *entries, int nr)
+{
+	int i, j;
+	unsigned int nr_zones = 0;
+	uint32_t zones[SD_MAX_REDUNDANCY];
+
+	for (i = 0; i < nr; i++) {
+		if (nr_zones >= ARRAY_SIZE(zones))
+			break;
+
+		for (j = 0; j < nr_zones; j++) {
+			if (zones[j] == entries[i].zone)
+				break;
+		}
+		if (j == nr_zones)
+			zones[nr_zones++] = entries[i].zone;
+	}
+
+	return min(sys->nr_sobjs, nr_zones);
+}
+
 /*
  * contains_node - checks that the node id is included in the target nodes
  *
@@ -1029,11 +1093,11 @@ uint64_t get_cluster_ctime(void)
  */
 static int contains_node(struct sheepdog_vnode_list_entry *key,
 			 struct sheepdog_vnode_list_entry *entry,
-			 int nr, int base_idx)
+			 int nr, int base_idx, int copies)
 {
 	int i;
 
-	for (i = 0; i < sys->nr_sobjs; i++) {
+	for (i = 0; i < copies; i++) {
 		int idx = get_nth_node(entry, nr, base_idx, i);
 		if (memcmp(key->addr, entry[idx].addr, sizeof(key->addr)) == 0
 		    && key->port == entry[idx].port)
@@ -1061,6 +1125,15 @@ struct recovery_work {
 	int nr_blocking;
 	int count;
 	uint64_t *oids;
+
+	int old_nr_nodes;
+	struct sheepdog_node_list_entry old_nodes[SD_MAX_NODES];
+	int cur_nr_nodes;
+	struct sheepdog_node_list_entry cur_nodes[SD_MAX_NODES];
+	int old_nr_vnodes;
+	struct sheepdog_vnode_list_entry old_vnodes[SD_MAX_VNODES];
+	int cur_nr_vnodes;
+	struct sheepdog_vnode_list_entry cur_vnodes[SD_MAX_VNODES];
 };
 
 static LIST_HEAD(recovery_work_list);
@@ -1098,41 +1171,42 @@ static struct recovery_work *recovering_work;
  * The node D, E, F, and A can recover objects from local, and the
  * node G recovers from the node B.
  */
-static int find_tgt_node(struct sheepdog_vnode_list_entry *old_entry, int old_nr, int old_idx,
-			 struct sheepdog_vnode_list_entry *cur_entry, int cur_nr, int cur_idx,
+static int find_tgt_node(struct sheepdog_vnode_list_entry *old_entry,
+			 int old_nr, int old_idx, int old_copies,
+			 struct sheepdog_vnode_list_entry *cur_entry,
+			 int cur_nr, int cur_idx, int cur_copies,
 			 int copy_idx)
 {
 	int i, j, idx;
 
-	dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n", old_idx, old_nr, cur_idx, cur_nr, copy_idx);
+	dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n",
+		old_idx, old_nr, old_copies, cur_idx, cur_nr, cur_copies, copy_idx);
 
-	if (copy_idx < sys->nr_sobjs) {
-		/* If the same node is in the previous target nodes, return its index */
-		idx = contains_node(cur_entry + get_nth_node(cur_entry, cur_nr, cur_idx, copy_idx),
-				    old_entry, old_nr, old_idx);
-		if (idx >= 0) {
-			dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n", idx, copy_idx, cur_idx, cur_nr);
-			return idx;
-		}
+	/* If the same node is in the previous target nodes, return its index */
+	idx = contains_node(cur_entry + get_nth_node(cur_entry, cur_nr, cur_idx, copy_idx),
+			    old_entry, old_nr, old_idx, old_copies);
+	if (idx >= 0) {
+		dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n", idx, copy_idx, cur_idx, cur_nr);
+		return idx;
 	}
 
 	for (i = 0, j = 0; ; i++, j++) {
 		if (i < copy_idx) {
 			/* Skip if the node can recover from its local */
 			idx = contains_node(cur_entry + get_nth_node(cur_entry, cur_nr, cur_idx, i),
-					    old_entry, old_nr, old_idx);
+					    old_entry, old_nr, old_idx, old_copies);
 			if (idx >= 0)
 				continue;
 
 			/* Find the next target which needs to recover from remote */
-			while (j < sys->nr_sobjs &&
+			while (j < old_copies &&
 			       contains_node(old_entry + get_nth_node(old_entry, old_nr, old_idx, j),
-					     cur_entry, cur_nr, cur_idx) >= 0)
+					     cur_entry, cur_nr, cur_idx, cur_copies) >= 0)
 				j++;
 		}
-		if (j == sys->nr_sobjs) {
+		if (j == old_copies) {
 			/*
-			 * Cannot find the target because the number of nodes
+			 * Cannot find the target because the number of zones
 			 * is smaller than the number of copies.  We can select
 			 * any node in this case, so select the first one.
 			 */
@@ -1153,8 +1227,10 @@ static int find_tgt_node(struct sheepdog_vnode_list_entry *old_entry, int old_nr
 }
 
 static int __recover_one(struct recovery_work *rw,
-			 struct sheepdog_vnode_list_entry *_old_entry, int old_nr,
-			 struct sheepdog_vnode_list_entry *_cur_entry, int cur_nr, int cur_idx,
+			 struct sheepdog_vnode_list_entry *_old_entry,
+			 int old_nr, int old_copies,
+			 struct sheepdog_vnode_list_entry *_cur_entry,
+			 int cur_nr, int cur_copies, int cur_idx,
 			 int copy_idx, uint32_t epoch, uint32_t tgt_epoch,
 			 uint64_t oid, char *buf, int buf_len)
 {
@@ -1164,22 +1240,30 @@ static int __recover_one(struct recovery_work *rw,
 	char name[128];
 	unsigned wlen = 0, rlen;
 	int fd, ret;
-	struct sheepdog_vnode_list_entry old_entry[SD_MAX_VNODES],
-		cur_entry[SD_MAX_VNODES], next_entry[SD_MAX_VNODES];
-	int next_nr;
+	struct sheepdog_vnode_list_entry *old_entry, *cur_entry, *next_entry;
+	int next_nr, next_copies;
 	int tgt_idx = -1;
 	int old_idx;
 
+	old_entry = malloc(sizeof(*old_entry) * SD_MAX_VNODES);
+	cur_entry = malloc(sizeof(*cur_entry) * SD_MAX_VNODES);
+	next_entry = malloc(sizeof(*next_entry) * SD_MAX_VNODES);
+	if (!old_entry || !cur_entry || !next_entry) {
+		eprintf("oom\n");
+		goto err;
+	}
+
 	memcpy(old_entry, _old_entry, sizeof(*old_entry) * old_nr);
 	memcpy(cur_entry, _cur_entry, sizeof(*cur_entry) * cur_nr);
 next:
 	dprintf("recover obj %"PRIx64" from epoch %"PRIu32"\n", oid, tgt_epoch);
 	old_idx = obj_to_sheep(old_entry, old_nr, oid, 0);
 
-	tgt_idx = find_tgt_node(old_entry, old_nr, old_idx, cur_entry, cur_nr, cur_idx, copy_idx);
+	tgt_idx = find_tgt_node(old_entry, old_nr, old_idx, old_copies,
+				cur_entry, cur_nr, cur_idx, cur_copies, copy_idx);
 	if (tgt_idx < 0) {
 		eprintf("cannot find target node, %"PRIx64"\n", oid);
-		return -1;
+		goto err;
 	}
 	e = old_entry + tgt_idx;
 
@@ -1192,22 +1276,24 @@ next:
 			 epoch, oid);
 		dprintf("link from %s to %s\n", old, new);
 		if (link(old, new) == 0)
-			return 0;
+			goto out;
 
 		if (errno == ENOENT) {
 			next_nr = epoch_log_read(tgt_epoch - 1, buf, buf_len);
 			if (next_nr <= 0) {
 				eprintf("no previous epoch, %"PRIu32"\n", tgt_epoch - 1);
-				return -1;
+				goto err;
 			}
+			next_nr /= sizeof(struct sheepdog_node_list_entry);
+			next_copies = get_max_copies((struct sheepdog_node_list_entry *)buf,
+						     next_nr);
 			next_nr = nodes_to_vnodes((struct sheepdog_node_list_entry *)buf,
-						  next_nr / sizeof(struct sheepdog_node_list_entry),
-						  next_entry);
+						  next_nr, next_entry);
 			goto not_found;
 		}
 
 		eprintf("cannot recover from local, %s, %s\n", old, new);
-		return -1;
+		goto err;
 	}
 
 	addr_to_str(name, sizeof(name), e->addr, 0);
@@ -1215,7 +1301,7 @@ next:
 	fd = connect_to(name, e->port);
 	if (fd < 0) {
 		eprintf("failed to connect to %s:%"PRIu32"\n", name, e->port);
-		return -1;
+		goto err;
 	}
 
 	if (is_vdi_obj(oid))
@@ -1237,9 +1323,9 @@ next:
 
 	close(fd);
 
-	if (ret < 0) {
+	if (ret != 0) {
 		eprintf("%"PRIu32"\n", rsp->result);
-		return -1;
+		goto err;
 	}
 
 	rsp = (struct sd_obj_rsp *)&hdr;
@@ -1256,20 +1342,20 @@ next:
 		fd = open(tmp_path, flags, def_fmode);
 		if (fd < 0) {
 			eprintf("failed to open %s, %s\n", tmp_path, strerror(errno));
-			return -1;
+			goto err;
 		}
 
 		ret = write(fd, buf, rlen);
 		if (ret != rlen) {
 			eprintf("failed to write object\n");
-			return -1;
+			goto err;
 		}
 
 		ret = fsetxattr(fd, ANAME_COPIES, &rsp->copies,
 				sizeof(rsp->copies), 0);
 		if (ret) {
 			eprintf("couldn't set xattr\n");
-			return -1;
+			goto err;
 		}
 
 		close(fd);
@@ -1278,47 +1364,60 @@ next:
 		ret = rename(tmp_path, path);
 		if (ret < 0) {
 			eprintf("failed to rename %s to %s, %m\n", tmp_path, path);
-			return -1;
+			goto err;
 		}
 		dprintf("recovered oid %"PRIx64" to epoch %"PRIu32"\n", oid, epoch);
-		return 0;
+		goto out;
 	}
 
 	if (rsp->result == SD_RES_NEW_NODE_VER || rsp->result == SD_RES_OLD_NODE_VER
 	    || rsp->result == SD_RES_NETWORK_ERROR) {
 		eprintf("try again, %"PRIu32", %"PRIx64"\n", rsp->result, oid);
 		rw->retry = 1;
-		return 0;
+		goto out;
 	}
 
 	if (rsp->result != SD_RES_NO_OBJ || rsp->data_length == 0) {
 		eprintf("%"PRIu32"\n", rsp->result);
-		return -1;
+		goto err;
 	}
+	next_nr = rsp->data_length / sizeof(struct sheepdog_node_list_entry);
+	next_copies = get_max_copies((struct sheepdog_node_list_entry *)buf, next_nr);
 	next_nr = nodes_to_vnodes((struct sheepdog_node_list_entry *)buf,
-				  rsp->data_length / sizeof(struct sheepdog_node_list_entry),
-				  next_entry);
+				  next_nr, next_entry);
 
 not_found:
-	for (copy_idx = 0; copy_idx < sys->nr_sobjs; copy_idx++)
+	for (copy_idx = 0; copy_idx < old_copies; copy_idx++)
 		if (get_nth_node(old_entry, old_nr, old_idx, copy_idx) == tgt_idx)
 			break;
-	if (copy_idx == sys->nr_sobjs) {
+	if (copy_idx == old_copies) {
 		eprintf("bug: cannot find the proper copy_idx\n");
-		return -1;
+		goto err;
 	}
 
 	dprintf("%"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32", %"PRIu32"\n", rsp->result, rsp->data_length, tgt_idx,
 		old_idx, old_nr, copy_idx);
 	memcpy(cur_entry, old_entry, sizeof(*old_entry) * old_nr);
+	cur_copies = old_copies;
 	cur_nr = old_nr;
 	cur_idx = old_idx;
 
 	memcpy(old_entry, next_entry, next_nr * sizeof(*next_entry));
+	old_copies = next_copies;
 	old_nr = next_nr;
 
 	tgt_epoch--;
 	goto next;
+out:
+	free(old_entry);
+	free(cur_entry);
+	free(next_entry);
+	return 0;
+err:
+	free(old_entry);
+	free(cur_entry);
+	free(next_entry);
+	return -1;
 }
 
 static void recover_one(struct work *work, int idx)
@@ -1327,15 +1426,10 @@ static void recover_one(struct work *work, int idx)
 	char *buf = NULL;
 	int ret;
 	uint64_t oid = rw->oids[rw->done];
-	struct sheepdog_node_list_entry old_nodes[SD_MAX_NODES];
-	struct sheepdog_node_list_entry cur_nodes[SD_MAX_NODES];
-	struct sheepdog_vnode_list_entry old_vnodes[SD_MAX_VNODES];
-	struct sheepdog_vnode_list_entry cur_vnodes[SD_MAX_VNODES];
-	int old_nr_nodes, cur_nr_nodes, old_nr_vnodes, cur_nr_vnodes;
+	int old_copies, cur_copies;
 	uint32_t epoch = rw->epoch;
 	int i, copy_idx = 0, cur_idx = -1;
 	int fd;
-	int nr_objs;
 
 	eprintf("%"PRIu32" %"PRIu32", %16"PRIx64"\n", rw->done, rw->count, oid);
 
@@ -1355,36 +1449,18 @@ static void recover_one(struct work *work, int idx)
 	else
 		buf = malloc(SD_DATA_OBJ_SIZE);
 
-	cur_nr_nodes = epoch_log_read(epoch, (char *)cur_nodes, sizeof(cur_nodes));
-	if (cur_nr_nodes <= 0) {
-		eprintf("failed to read current epoch, %"PRIu32"\n", epoch);
-		goto out;
-	}
-	cur_nr_nodes /= sizeof(struct sheepdog_node_list_entry);
-
-	old_nr_nodes = epoch_log_read(epoch - 1, (char *)old_nodes, sizeof(old_nodes));
-	if (old_nr_nodes <= 0) {
-		eprintf("failed to read previous epoch, %"PRIu32"\n", epoch - 1);
-		goto fail;
-	}
-	old_nr_nodes /= sizeof(struct sheepdog_node_list_entry);
-
-	old_nr_vnodes = nodes_to_vnodes(old_nodes, old_nr_nodes, old_vnodes);
-	cur_nr_vnodes = nodes_to_vnodes(cur_nodes, cur_nr_nodes, cur_vnodes);
-
 	if (!sys->nr_sobjs)
 		goto fail;
 
-	cur_idx = obj_to_sheep(cur_vnodes, cur_nr_vnodes, oid, 0);
+	cur_idx = obj_to_sheep(rw->cur_vnodes, rw->cur_nr_vnodes, oid, 0);
 
-	nr_objs = sys->nr_sobjs;
-	if (nr_objs > cur_nr_nodes)
-		nr_objs = cur_nr_nodes;
+	old_copies = get_max_copies(rw->old_nodes, rw->old_nr_nodes);
+	cur_copies = get_max_copies(rw->cur_nodes, rw->cur_nr_nodes);
 
 	copy_idx = -1;
-	for (i = 0; i < nr_objs; i++) {
-		int n = obj_to_sheep(cur_vnodes, cur_nr_vnodes, oid, i);
-		if (is_myself(cur_vnodes[n].addr, cur_vnodes[n].port)) {
+	for (i = 0; i < cur_copies; i++) {
+		int n = obj_to_sheep(rw->cur_vnodes, rw->cur_nr_vnodes, oid, i);
+		if (is_myself(rw->cur_vnodes[n].addr, rw->cur_vnodes[n].port)) {
 			copy_idx = i;
 			break;
 		}
@@ -1394,19 +1470,21 @@ static void recover_one(struct work *work, int idx)
 		goto out;
 	}
 
-	dprintf("%"PRIu32", %"PRIu32", %"PRIu32"\n", cur_idx, cur_nr_nodes, copy_idx);
+	dprintf("%"PRIu32", %"PRIu32", %"PRIu32"\n", cur_idx, rw->cur_nr_nodes,
+		copy_idx);
 
-	ret = __recover_one(rw, old_vnodes, old_nr_vnodes, cur_vnodes, cur_nr_vnodes,
+	ret = __recover_one(rw, rw->old_vnodes, rw->old_nr_vnodes, old_copies,
+			    rw->cur_vnodes, rw->cur_nr_vnodes, cur_copies,
 			    cur_idx, copy_idx, epoch, epoch - 1, oid,
 			    buf, SD_DATA_OBJ_SIZE);
 	if (ret == 0)
 		goto out;
 
-	for (i = 0; i < sys->nr_sobjs; i++) {
+	for (i = 0; i < cur_copies; i++) {
 		if (i == copy_idx)
 			continue;
-		ret = __recover_one(rw, old_vnodes, old_nr_vnodes,
-				    cur_vnodes, cur_nr_vnodes, cur_idx, i,
+		ret = __recover_one(rw, rw->old_vnodes, rw->old_nr_vnodes, old_copies,
+				    rw->cur_vnodes, rw->cur_nr_vnodes, cur_copies, cur_idx, i,
 				    epoch, epoch - 1, oid, buf, SD_DATA_OBJ_SIZE);
 		if (ret == 0)
 			goto out;
@@ -1414,8 +1492,7 @@ static void recover_one(struct work *work, int idx)
 fail:
 	eprintf("failed to recover object %"PRIx64"\n", oid);
 out:
-	if (buf)
-		free(buf);
+	free(buf);
 }
 
 static struct recovery_work *suspended_recovery_work;
@@ -1432,7 +1509,7 @@ static void recover_timer(void *data)
 		return;
 	}
 
-	queue_work(&rw->work);
+	queue_work(sys->recovery_wqueue, &rw->work);
 }
 
 void resume_recovery_work(void)
@@ -1450,7 +1527,7 @@ void resume_recovery_work(void)
 		return;
 
 	suspended_recovery_work = NULL;
-	queue_work(&rw->work);
+	queue_work(sys->recovery_wqueue, &rw->work);
 }
 
 int is_recoverying_oid(uint64_t oid)
@@ -1542,7 +1619,7 @@ static void recover_done(struct work *work, int idx)
 			return;
 		}
 		resume_pending_requests();
-		queue_work(&rw->work);
+		queue_work(sys->recovery_wqueue, &rw->work);
 		return;
 	}
 
@@ -1562,7 +1639,7 @@ static void recover_done(struct work *work, int idx)
 		list_del(&rw->rw_siblings);
 
 		recovering_work = rw;
-		queue_work(&rw->work);
+		queue_work(sys->recovery_wqueue, &rw->work);
 	}
 }
 
@@ -1590,8 +1667,6 @@ static int __fill_obj_list(struct sheepdog_node_list_entry *e, uint32_t epoch,
 
 	memset(&hdr, 0, sizeof(hdr));
 	hdr.opcode = SD_OP_GET_OBJ_LIST;
-	/* we don't need to set epoch */
-	hdr.epoch = epoch;
 	hdr.tgt_epoch = epoch - 1;
 	hdr.flags = 0;
 	hdr.data_length = rlen;
@@ -1617,17 +1692,20 @@ static int merge_objlist(struct sheepdog_vnode_list_entry *entries, int nr_entri
 			 uint64_t *list2, int nr_list2, int nr_objs)
 {
 	int i, j, idx;
+	int old_nr_list1 = nr_list1;
 
 	for (i = 0; i < nr_list2; i++) {
-		for (j = 0; j < nr_objs; j++) {
-			idx = obj_to_sheep(entries, nr_entries, list2[i], j);
-			if (is_myself(entries[idx].addr, entries[idx].port))
-				break;
+		if (entries) {
+			for (j = 0; j < nr_objs; j++) {
+				idx = obj_to_sheep(entries, nr_entries, list2[i], j);
+				if (is_myself(entries[idx].addr, entries[idx].port))
+					break;
+			}
+			if (j == nr_objs)
+				continue;
 		}
-		if (j == nr_objs)
-			continue;
 
-		if (bsearch(list2 + i, list1, nr_list1, sizeof(*list1), obj_cmp))
+		if (bsearch(list2 + i, list1, old_nr_list1, sizeof(*list1), obj_cmp))
 			continue;
 
 		list1[nr_list1++] = list2[i];
@@ -1648,11 +1726,12 @@ static int fill_obj_list(struct recovery_work *rw,
 	int i, j;
 	uint8_t *buf = NULL;
 	size_t buf_size = SD_DATA_OBJ_SIZE; /* FIXME */
-	struct sheepdog_vnode_list_entry vnodes[SD_MAX_VNODES];
+	struct sheepdog_vnode_list_entry *vnodes;
 	int nr_vnodes, retry_cnt = 0;
 
+	vnodes = malloc(sizeof(*vnodes) * SD_MAX_VNODES);
 	buf = malloc(buf_size);
-	if (!buf)
+	if (!buf || !vnodes)
 		goto fail;
 
 	nr_vnodes = nodes_to_vnodes(cur_entry, cur_nr, vnodes);
@@ -1685,9 +1764,11 @@ static int fill_obj_list(struct recovery_work *rw,
 					  rw->count, (uint64_t *)buf, nr, nr_objs);
 	}
 
+	free(vnodes);
 	free(buf);
 	return 0;
 fail:
+	free(vnodes);
 	free(buf);
 	rw->retry = 1;
 	return -1;
@@ -1697,67 +1778,44 @@ static void __start_recovery(struct work *work, int idx)
 {
 	struct recovery_work *rw = container_of(work, struct recovery_work, work);
 	uint32_t epoch = rw->epoch;
-	struct sheepdog_node_list_entry old_entry[SD_MAX_NODES],
-		cur_entry[SD_MAX_NODES];
-	int old_nr, cur_nr, nr_objs;
-	int fd;
-	char path[PATH_MAX], tmp_path[PATH_MAX];
-	int ret;
+	int nr_objs;
 
 	dprintf("%u\n", epoch);
 
-	cur_nr = epoch_log_read(epoch, (char *)cur_entry, sizeof(cur_entry));
-	if (cur_nr <= 0) {
-		eprintf("failed to read epoch log, %"PRIu32"\n", epoch);
-		goto fail;
+	if (rw->cur_nr_nodes == 0) {
+		/* setup node list and virtual node list */
+		rw->cur_nr_nodes = epoch_log_read(epoch, (char *)rw->cur_nodes,
+						  sizeof(rw->cur_nodes));
+		if (rw->cur_nr_nodes <= 0) {
+			eprintf("failed to read epoch log, %"PRIu32"\n", epoch);
+			goto fail;
+		}
+		rw->cur_nr_nodes /= sizeof(struct sheepdog_node_list_entry);
+
+		rw->old_nr_nodes = epoch_log_read(epoch - 1, (char *)rw->old_nodes,
+						  sizeof(rw->old_nodes));
+		if (rw->old_nr_nodes <= 0) {
+			eprintf("failed to read epoch log, %"PRIu32"\n", epoch - 1);
+			goto fail;
+		}
+		rw->old_nr_nodes /= sizeof(struct sheepdog_node_list_entry);
+
+		rw->old_nr_vnodes = nodes_to_vnodes(rw->old_nodes, rw->old_nr_nodes,
+						    rw->old_vnodes);
+		rw->cur_nr_vnodes = nodes_to_vnodes(rw->cur_nodes, rw->cur_nr_nodes,
+						    rw->cur_vnodes);
 	}
-	cur_nr /= sizeof(struct sheepdog_node_list_entry);
 
-	old_nr = epoch_log_read(epoch - 1, (char *)old_entry, sizeof(old_entry));
-	if (old_nr <= 0) {
-		eprintf("failed to read epoch log, %"PRIu32"\n", epoch - 1);
+	if (!sys->nr_sobjs)
 		goto fail;
-	}
-	old_nr /= sizeof(struct sheepdog_node_list_entry);
+	nr_objs = get_max_copies(rw->cur_nodes, rw->cur_nr_nodes);
 
-	nr_objs = sys->nr_sobjs;
-	if (nr_objs > cur_nr)
-		nr_objs = cur_nr;
-	if (!nr_objs)
-		goto fail;
-
-	if (fill_obj_list(rw, old_entry, old_nr, cur_entry, cur_nr, nr_objs) != 0) {
+	if (fill_obj_list(rw, rw->old_nodes, rw->old_nr_nodes, rw->cur_nodes,
+			  rw->cur_nr_nodes, nr_objs) != 0) {
 		eprintf("fatal recovery error\n");
 		goto fail;
 	}
 
-	if (rw->retry)
-		goto fail;
-
-	snprintf(path, sizeof(path), "%s%08u/list", obj_path, epoch);
-	snprintf(tmp_path, sizeof(tmp_path), "%s%08u/list.tmp", obj_path, epoch);
-
-	dprintf("write object list to %s\n", tmp_path);
-	fd = open(tmp_path, O_RDWR | O_CREAT | O_SYNC, def_fmode);
-	if (fd < 0) {
-		eprintf("failed to open %s, %s, %m\n", tmp_path, strerror(errno));
-		goto fail;
-	}
-	ret = write(fd, rw->oids, sizeof(*rw->oids) * rw->count);
-	if (ret != sizeof(uint64_t) * rw->count) {
-		eprintf("failed to write to %s, %m\n", tmp_path);
-		close(fd);
-		goto fail;
-	}
-	close(fd);
-
-	dprintf("rename %s to %s\n", tmp_path, path);
-	ret = rename(tmp_path, path);
-	if (ret < 0) {
-		eprintf("failed to rename %s to %s, %m\n", tmp_path, path);
-		goto fail;
-	}
-
 	return;
 fail:
 	rw->count = 0;
@@ -1784,7 +1842,7 @@ int start_recovery(uint32_t epoch)
 		list_add_tail(&rw->rw_siblings, &recovery_work_list);
 	else {
 		recovering_work = rw;
-		queue_work(&rw->work);
+		queue_work(sys->recovery_wqueue, &rw->work);
 	}
 
 	return 0;
@@ -1852,7 +1910,7 @@ static int attr(char *path, const char *name, void *var, int len, int set)
 	return SD_RES_SUCCESS;
 }
 
-static int init_base_path(const char *d)
+int init_base_path(const char *d)
 {
 	int new = 0;
 
@@ -1876,7 +1934,7 @@ static int init_obj_path(const char *base_path)
 static int init_epoch_path(const char *base_path)
 {
 	int new, ret;
-	uint32_t epoch;
+	uint32_t epoch, latest_epoch;
 	DIR *dir;
 	char path[1024];
 	struct dirent *dent;
@@ -1889,38 +1947,39 @@ static int init_epoch_path(const char *base_path)
 	if (new || ret)
 		return ret;
 
-	epoch = get_latest_epoch();
+	latest_epoch = get_latest_epoch();
 
-	snprintf(path, sizeof(path), "%s/%08u", obj_path, epoch);
+	for (epoch = 1; epoch <= latest_epoch; epoch++) {
+		snprintf(path, sizeof(path), "%s/%08u", obj_path, epoch);
 
-	vprintf(SDOG_INFO "found the epoch dir, %s\n", path);
+		vprintf(SDOG_INFO "found the obj dir, %s\n", path);
 
-	dir = opendir(path);
-	if (!dir) {
-		if (errno == ENOENT)
-			/* probably, sheepdog is not formatted yet  */
-			return 0;
+		dir = opendir(path);
+		if (!dir) {
+			if (errno == ENOENT)
+				continue;
 
-		vprintf(SDOG_ERR "failed to open the epoch dir, %m\n");
-		return SD_RES_EIO;
+			vprintf(SDOG_ERR "failed to open the epoch dir, %m\n");
+			return SD_RES_EIO;
+		}
+
+		while ((dent = readdir(dir))) {
+			if (!strcmp(dent->d_name, ".") ||
+			    !strcmp(dent->d_name, ".."))
+				continue;
+
+			oid = strtoull(dent->d_name, NULL, 16);
+
+			if (!is_vdi_obj(oid))
+				continue;
+
+			vprintf(SDOG_DEBUG "found the vdi obj, %" PRIx64 "\n", oid);
+
+			set_bit(oid_to_vid(oid), sys->vdi_inuse);
+		}
+		closedir(dir);
 	}
 
-	while ((dent = readdir(dir))) {
-		if (!strcmp(dent->d_name, ".") ||
-		    !strcmp(dent->d_name, ".."))
-			continue;
-
-		oid = strtoull(dent->d_name, NULL, 16);
-
-		if (!is_vdi_obj(oid))
-			continue;
-
-		vprintf(SDOG_DEBUG "found the vdi obj, %" PRIx64 "\n", oid);
-
-		set_bit(oid_to_vid(oid), sys->vdi_inuse);
-	}
-	closedir(dir);
-
 	return 0;
 }
 
@@ -1980,10 +2039,6 @@ int init_store(const char *d)
 {
 	int ret;
 
-	ret = init_base_path(d);
-	if (ret)
-		return ret;
-
 	ret = init_obj_path(d);
 	if (ret)
 		return ret;
@@ -2162,6 +2217,8 @@ inline int jrnl_apply_to_target_object(struct jrnl_file_desc *jfd)
 	uint32_t jrnl_type;
 
 	ret = jrnl_get_type_from_file(jfd, &jrnl_type);
+	if (ret)
+		return ret;
 
 	return jrnl_handlers[jrnl_type].apply_to_target_object(jfd);
 }
